/* %W% %G%

Tom's incredible emacs debugger interface.

You shouldn't need to read this file. Just loading it will turn it on.

*/

:- module(tdb, [tdb_on/0,tdb_off/0,tdb_reset/0]).

:- use_module(library(avl)).
:- use_module(library(between)).
:- use_module(library(directory)).
:- use_module(library(environ)).
:- use_module(library(fparse)).
:- use_module(library(strings)).

:- dynamic tdb_enabled/0.	% true when source linked debugging is
				% turned on.

:- dynamic under_emacs/0.	% true when this prolog session is a
				% slave to an emacs process

% last_update(Last) stores the debugger level corresponding with the
% most recently displayed source.

:- dynamic last_update/1.

% q(Level, File, Beginning, End, BreakLevel) is a way of storing the
% offsets that will display the source corresponding to the debugger
% level.

:- dynamic q/5.

% f(Filname, ModifyDate, Avl) is used to store the modify date of the
% file and its associated avl tree.

:- dynamic f/3, break_level/1, debug_tdb/0, error/3.

break_level(0).

% debugging this stuff is scary, so instead I use this wrapper

dbg(G) :-
   (  debug_tdb
   -> format(user_error, '~N~2ncalling ~q~n', [G]),
      (  G
      -> format(user_error, '~N~2ndone ~q~n', [G])
      ;  format(user_error, '~N~2nfail ~q~n', [G])
      ),
      nl
   ;  G
   ).

% tdb_on/0 and tdb_off/0 are commands that turn on and off source linked
% debugging.

tdb_on :- 
    (   tdb_enabled
    ->  true
    ;   assert(tdb_enabled)
    ).
tdb_off :- retractall(tdb_enabled).

tdb_reset :- retractall(f(_,_,_)), retractall(q(_,_,_,_,_)),
    retractall(last_update(_)), retractall(break_level(_)).

% eek/1 is a generate_message_hook. It traps debugger messages that
% require source file look ups. If it is a "head" message, then display
% the source, otherwise if it is returning to a previous level, display
% the source for the new level.

eek(_) :- \+ tdb_enabled, !.

eek(X) :-

   debug_tdb, current_output(CO), set_output(user_error), nl,
   portray_clause(X), nl, set_output(CO),
   fail.

eek(_) :-

   retract(error(C, F, B)), character_count(user, E),
   to_emacs(['tdb-compilation-error ',C,' "',F,'" ',B,' ',E]),
   fail.

eek(debugger(_Style,_Err,_Prefix,Invoc,Depth,Extra,Port,Goal,_PDep,_Flag)) :-

   tdb_enabled,

   highlight_ancestor(Port, Invoc),

   (   clause_number(Extra, CN)
   ->

       % if we've already fetched the positional information for this
       % file and it is current, then we're ready, otherwise possibly
       % clean up and then fetch the positional information.

       m(Goal, user, G, Mod), (source_file(Mod:G, CN, F) -> true),
       file_property(F, modify_time, M),
       (  f(F, M, P)
       -> true
       ;  retractall(f(F,_,_)), read_terms(F, P), assert(f(F, M, P))
       ),
       functor(G, N, A), avl_fetch(g(N,A,CN), P, [Beg|End]),
       asserta(q(Depth, F, Beg, End, BL))
   ;   

       % discard all obviated q/5 structures, then if there are some
       % remaining sources and the most recently displayed source is
       % not being displayed, display it

       (   q(L, _, _, _, BL), L > Depth, retract(q(L,_,_,_,BL)), fail
       ;   \+last_update(Depth), q(Depth, F, Beg, End, BL)
       )
   ),

   display_source(F, Beg, End),
   retractall(last_update(_)), assert(last_update(Depth)).

% make sure we get rid of junk from old break levels after an abort

eek(top_level(_,N,_)) :- to_emacs(['tdb-undo']), break_level(Old),
    set_break_level(N), retractall(last_update(_)),
    M is N+1, between(M, Old, X), retractall(q(_,_,_,_,X)).

% handle a graceful return to the debugger from a break level

eek(end_break_level(N)) :-
   BL is N-1, set_break_level(BL).

% ensure that extents get written at the right place since prolog's
% notion of the stream position is often in conflict with emacs.

eek(debugger(standard_help)) :- reset_offset.
eek(term_reading) :- reset_offset.

% capture errors

eek(error_at(pos_and_file('$stream_position'(C,_,_,_,_),F))) :-
   capture_errors(C, F).
eek(syntax_error(_,between('$stream_position'(C,_,_,_,_),_),_,_,_,F)) :-
   capture_errors(C, F).

% determine the module and goal

m(M0:G0, _, G, M) :- !, m(G0, M0, G, M).
m(G, M, G, M).

% fetch the clause number from the "Extra" structure. Fail if it is not
% a head/[1,2] structure.

clause_number(head(CN), CN).
clause_number(head(CN,_), CN).

% create an avl tree that has as its keys a g(N,A,C) structure and as
% its values a [B|E] structure, where N is the clause Name, A is its
% arity, C is its clause number, B is the beginning position in the
% file, and E is its ending position.

read_terms(File, Avl) :-
    see(File), to_emacs(['message "tdb: reading ',File,'"']),
    prolog_flag(syntax_errors, Old, quiet),
    rt(foo, G, Position, Term), rts(Term, G, Position, empty, Avl),
    prolog_flag(syntax_errors, _, Old), seen, to_emacs(['message nil']).

% fetch a clause head and its term position from its source file

rt(Last, G, Position, T) :-
   (  read_term([subterm_positions(STP)], T0), expand_term(T0, T1),
      rt1(T1, STP, Last, G, Position, T)
   -> true
   ;  rt(Last, G, Position, T)
   ).

rt1(end_of_file, _, _, _, _, end_of_file) :- !.
rt1(T1, STP, Last, G, Position, T) :-
   head(T1, T), subterm_position(STP, Position),
   functor(T, N, A), last(Last, N, A, G).
   
% isolate clause heads, discarding compiler directives

head((:-_), _) :- !, fail.
head((T:-_), T) :- !.
head(T, T).

% find the beginning and end of a clause given its term position

subterm_position(term_position(B,E,_,_,_), [B|E]).
subterm_position(B-E, [B|E]).

% number clauses

last(g(N,A,C0), N, A, g(N,A,C)) :- !, C is C0+1.
last(_, N, A, g(N,A,1)).

% consume terms from a file by placing each clause position in an avl tree

rts(end_of_file, _, _, Avl, Avl) :- !.
rts(_, G, L, Avl0, Avl) :-
    avl_store(G, Avl0, L, Avl1), rt(G, G1, L1, T), rts(T, G1, L1, Avl1, Avl).

% send lisp to emacs.

to_emacs(List) :- current_output(CO),
    (   under_emacs
    ->  open('/dev/tty', write, [binary], Stream), set_output(Stream), put(30)
    ;   topen(Stream), set_output(Stream)
    ),
    put(0'(), write_list(List), put(0')),
    (   under_emacs
    ->  put(29), close(Stream)
    ;   tclose(Stream)
    ),
    set_output(CO),
    (   debug_tdb
    ->  print(user_error, tdb(List))
    ;   true
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% only load the tcp stuff if we absolutely have to.

:- dynamic topen/1, tclose/1.

topen(_) :-
   abolish([topen/1,tclose/1]),
   source_file(tdb_on, F), fparse(F, Dir, _, _),
   concat_atom([Dir,/,topen], Topen),
   use_module(Topen).

tclose(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_list([]).
write_list([H|T]) :- write(H), write_list(T).

% record the current break level

set_break_level(N) :- retractall(break_level(_)), assert(break_level(N)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% highlight ancestors

% last_ancestor(CharacterPosition, InvocationNumber) ... for
% highlighting ancester entries for retry

:- dynamic last_ancestor/2.

highlight_ancestor('', Invoc) :- !,
   cc(E), 
   (  retract(last_ancestor(B, Invoc0))
   -> dbg(highlight(B, E, Invoc0))
   ;  true
   ),
   dbg(assert(last_ancestor(E, Invoc))).
highlight_ancestor(_, _) :-
   (  retract(last_ancestor(B, Invoc))
   -> cc(E), dbg(highlight(B, E, Invoc))
   ;  true
   ).

cc(E) :-
   flush_output(user_output), flush_output(user_error),
   character_count(user_output, E).

highlight(B0, E, Invoc) :-
   B is B0+1, to_emacs(['tdb-ancestor ',B,' ',E,' ',Invoc]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

reset_offset :-
   character_count(user, C), to_emacs(['tdb-character-count ',C]).

capture_errors(C, File) :-
   character_count(user, B), assert(error(C, File, B)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

display_source(F, Beg, End) :-
   to_emacs(['tdb "',F,'" ',Beg,' ',End]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% this clause must be last

:- multifile user:generate_message_hook/3.

user:generate_message_hook(X,_,_) :- eek(X), fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% intialization ... maybe this should not be in this file

:- initialization
    source_file(tdb_on, F), fparse(F, Dir, Name, _),
    (  environ('TERM', emacs)
    -> assert(under_emacs),
       to_emacs(['if (not (boundp ''tdb-was-loaded)) (load "',Dir,Name,'")']),
       tdb_on
    ;  true
    ).

end_of_file.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% debugging stuff

debug_tdb.

:- use_module(library(addportray)).

my_avl_portrayal(node(_,_,_,_,_)) :- write(avl).

:- initialization add_portray(my_avl_portrayal).

tdb_portrayal(tdb(List)) :- nl, put("("), write_list(List), put(")"), nl.

:- initialization add_portray(tdb_portrayal).
